#!/usr/bin/env python2
#
# Usage: ./fungewars.py [-n] <program> [<time>]
# -n = don't use curses interface (runs much faster)
# <program> = Befunge program to load
# <time> = number of milliseconds to wait between steps
#
# Copyright (c) 2001 Rob Speer
# This program is released under the MIT license. See LICENSE for details.

import traceback, os, string, time
from curses import *
from random import *
from struct import *
from sys import *

debug = 0
fungesize = (80, 25)
progs = 2

max_stack = 1000

ver = 45 # version 0.4.5 (FungeWars)

def chomp(string):
	if string[-1] == '\012': return string[:-1]
	else: return string

def output_text(string):
	global output
	if string != '\000': 
		if string >= " " or string == '\012': output += string
		else: output += "?"
	if not curses: stdout.write(string)

def buffer_read():
	global output
	if len(output) < 1: return ''
	else:
		a = output[0]
		output = output[1:]
		return a
	
def findsize(source):
	width = 0
	for line in source:
		if len(line) > width:
			width = len(line)
	return (width, len(source))
		

class ipteam:
	def __init__(self,firstip):
		self.iplist = [firstip]
		self.turn = 0
	def add_ip(self,ip):
		self.iplist[self.turn:self.turn] = [ip]
	def del_ip(self,ip):
		del self.iplist[self.iplist.index(ip)]
		self.turn -= 1
	def next_ip(self):
		self.turn += 1
		self.turn %= len(self.iplist)
		return self.iplist[self.turn]
	def has_ip(self,ip):
		return (ip in self.iplist)
	def members(self):
		return len(self.iplist)

class fungespace:
	def __init__(self,x,y):
		self.size = (x,y)
		self.grid = [["z"]*x for i in range(y)]
	def __getitem__(self,y):
		return self.grid[y % (self.size[1])]
#	def movestart(self,newstartx,newstarty):
#		for row in self.grid:
#			row.movestart(newstartx)
#		a = self.start[0] - newstartx
#		b = self.start[1] - newstarty
#		self.grid.insert(0,[fungerow(self.size[0] + a, newstartx) for i in range(b)])
#		self.start = (newstartx,newstarty)
#		self.size = (self.size[0] + a, self.size[1] + b)
#	def resize(self,newx,newy):
#		for row in self.grid:
#			row.resize(newx)
#		b = newy - self.size[1]
#		self.grid.extend([fungerow(newx, self.start[0]) for i in range(b)])
#		self.size = (newx, newy)
	def putgrid(self,x,y,char):
		self.grid[y % (self.size[1])][x % (self.size[0])] = char
	def loadat(self,source,x_y,width_height):
		x, y = x_y
		width, height = width_height
		for linenum in range(height):
			for charnum in range(width):
				if charnum >= len(source[linenum]):
					self.putgrid(x+charnum,y+linenum,' ')
				else:
					self.putgrid(x+charnum,y+linenum,source[linenum][charnum])



class ip:
	def __init__(self, xcoord, ycoord, xdelta, ydelta, xoffset, yoffset, team):
		global uniqueid
		self.x = xcoord
		self.y = ycoord
		self.dx = xdelta
		self.dy = ydelta
		self.sx = xoffset
		self.sy = yoffset
		self.stack = [[]]
		uniqueid += 1
		self.id = uniqueid
		self.stringmode = 0
		self.team = team
	def setdelta(self, xdelta, ydelta):
		self.dx = xdelta
		self.dy = ydelta
	def setpos(self, xcoord, ycoord):
		self.x = xcoord
		self.y = ycoord
	def setstore(self, xoffset, yoffset):
		self.sx = xoffset
		self.sy = yoffset
	def info(self):
		greatestptx = greatestpty = leastptx = leastpty = 0
		#infolist = [9, # change to 15 when 'i' and 'o' are implemented
		#999999999,     # number of bytes per cell - good approximation to infinity
		#0x50794675,    # Handprint: "PyFu"
		#ver,           # version number
		#1,             # system calls with = go to system()
		#ord(os.sep),   # separator character
		#2,             # 2 dimensions (Befunge)
		#self.id,       # Pointer ID
		#self.team,     # team number
		infolist = [9, 999999999, 0x50794675, ver, 1, ord(os.sep), 2,\
		self.id, self.team, self.x, self.y, self.dx, self.dy, self.sx,\
		self.sy, leastptx, leastpty, greatestptx, greatestpty,\
		((time.localtime(0)[0]-1900)*65536 + (time.localtime(0)[1])*256 +\
		time.localtime(0)[2]), ((time.localtime(0)[3])*65536 +\
		(time.localtime(0)[4])*256 + time.localtime(0)[5])]
		for i in argv: # command-line arguments
			for char in i:
				infolist.append(ord(char))
			infolist.append(0)
		infolist.extend([0,0])
		for i in os.environ.keys(): # environment variables
			s = i+'='+os.environ[i]
			for char in s:
				infolist.append(ord(char))
			infolist.append(0)
		infolist.append(0)
		return infolist[:max_stack-50]
	def popstring(self):
		s = ''
		a = self.spop(0)
		while (len(self.stack[0]) > 0) and (a != 0):
			s += chr(a%256)
			a = self.spop(0)
		return s
	def movewrap(self): # only torus wrapping in FungeWars
		self.x += self.dx
		self.x %= fungesize[0] 
		self.y += self.dy
		self.y %= fungesize[1]
	def findnext(self):
		global program
		start = (self.x, self.y)
		s = program[self.y][self.x]
		self.movewrap()
		if (self.stringmode == 0) or (s == ' '):
			while (program[self.y][self.x] == ' ') and\
			(self.x, self.y) != start:
				self.movewrap()
		if (self.stringmode == 0):
			if program[self.y][self.x] == ';':
				self.movewrap()
				while program[self.y][self.x] != ';':
					self.movewrap()
				self.findnext()
	def spop(self,n):
		try:
			a = self.stack[n].pop()
		except:
			a = 0
		return a


def run(c,inst): # c is the current instruction pointer
	global program, ipteams
	if c.stringmode == 0:
		if inst == '>': # go east
			c.setdelta(1,0)
		elif inst == '<': # go west
			c.setdelta(-1,0)
		elif inst == '^': # go north
			c.setdelta(0,-1)
		elif inst == 'v': # go south
			c.setdelta(0,1)
		elif inst == '?': # go away
			a = randint(0,3)
			b = [(1,0),(-1,0),(0,-1),(0,1)][a]
			c.setdelta(b[0],b[1])
		elif inst == '[': # turn left
			c.setdelta(c.dy, -c.dx)
		elif inst == ']': # turn right
			c.setdelta(-c.dy, c.dx)
		elif inst == 'r': # reverse
			c.setdelta(-c.dx, -c.dy)
		elif inst == 'x': # absolute vector
			b = c.spop(0)
			a = c.spop(0)
			c.setdelta(a, b)
		elif inst == 'z': # nop
			pass
		elif inst == '#': # trampoline
			#c.x += c.dx
			#c.y += c.dy
			c.movewrap()
		elif inst == '@' or inst == 'q' or inst == '\0': # die
			team = ipteams[c.team]
			team.del_ip(c)
			del iplist[iplist.index(c)]
			if team.members() == 0:
				if c.team == 0: exit("Green player wins.")
				else: exit("Red player wins.")
		elif inst == 'j': # jump over
			a = c.spop(0)
			#c.x += a * c.dx
			#c.y += a * c.dy
			for i in range(a):
				c.movewrap()
		elif inst == 'k': # iterate
			c.findnext()
			newinst = program[c.y][c.x]
			b = c.spop(0)
			for i in range(b):
				run(c,newinst)
		elif inst == '!': # logical not
			a = c.spop(0)
			if a:
				c.stack[0].append(0)
			else:
				c.stack[0].append(1)
		elif inst == '`': # greater than
			b = c.spop(0)
			a = c.spop(0)
			if a > b:
				c.stack[0].append(1)
			else:
				c.stack[0].append(0)
		elif inst == '_': # east-west if
			a = c.spop(0)
			if a:
				c.setdelta(-1,0)
			else:
				c.setdelta(1,0)
		elif inst == '|': # north-south if
			a = c.spop(0)
			if a:
				c.setdelta(0,-1)
			else:
				c.setdelta(0,1)
		elif inst == 'w': # compare and turn
			b = c.spop(0)
			a = c.spop(0)
			if a < b:
				c.setdelta(c.dy, -c.dx)
			elif a > b:
				c.setdelta(-c.dy, c.dx)
		elif ('0' <= inst) and (inst <= '9'): # push 0-9
			c.stack[0].append(int(inst))
		elif inst == 'a': # push ten
			c.stack[0].append(10)
		elif inst == 'b': # push eleven
			c.stack[0].append(11)
		elif inst == 'c': # push twelve
			c.stack[0].append(12)
		elif inst == 'd': # push thirteen
			c.stack[0].append(13)
		elif inst == 'e': # push fourteen
			c.stack[0].append(14)
		elif inst == 'f': # push fifteen
			c.stack[0].append(15)
		elif inst == '+': # add
			a = c.spop(0)
			b = c.spop(0)
			c.stack[0].append(long(a+b))
		elif inst == '-': # subtract
			a = c.spop(0)
			b = c.spop(0)
			c.stack[0].append(long(b-a))
		elif inst == '*': # multiply
			a = c.spop(0)
			b = c.spop(0)
			c.stack[0].append(long(a*b))
		elif inst == '/': # divide
			a = c.spop(0)
			b = c.spop(0)
			if a == 0:
				c.stack[0].append(0)
			else:
				c.stack[0].append(long(b/a))
		elif inst == '%': # modulo
			a = c.spop(0)
			b = c.spop(0)
			if a == 0:
				c.stack[0].append(0)
			else:
				c.stack[0].append(long(b%a))
		elif inst == '"': # start stringmode
			c.stringmode = 1
		elif inst == "'": # fetch character
			#c.x += c.dx
			#c.y += c.dy
			c.movewrap()
			s = ord(program[c.y][c.x])
			c.stack[0].append(s)
		elif inst == "s": # store character
			b = c.y + c.dy
			a = c.x + c.dx
			s = chr(c.spop(0)%256)
			program.putgrid(a,b,s)
			if curses:
				mungechar(b % fungesize[1],\
				a % fungesize[0], s, A_BOLD)
			
		elif inst == "$": # pop
			if len(c.stack[0]) > 0: del c.stack[0][-1]
		elif inst == ":": # duplicate
			a = c.spop(0)
			c.stack[0].extend([a,a])
		elif inst == "\\": # swap
			a = c.spop(0)
			b = c.spop(0)
			c.stack[0].extend([a,b])
		elif inst == "n": # stack clear
			c.stack[0] = []
		elif inst == "{": # begin block
			n = c.spop(0)
			c.stack.insert(0,[])
			if n > 0:
				if len(c.stack[1]) < n:
					c.stack[0] = c.stack[1] + [0]*(n - len(c.stack[1]))
					c.stack[1] = []
				else:
					c.stack[0] = c.stack[1][-n:]
					c.stack[1][-n:] = []
			elif n < 0:
				c.stack[1].extend([0]*abs(n))
			c.stack[1].extend([c.sx,c.sy])
			c.setstore(c.x + c.dx, c.y + c.dy)
		elif inst == "}": # end block
			if len(c.stack) < 2:
				c.setdelta(-c.dx,-c.dy) # reflect on failure
			else:
				n = c.spop(0)
				b = c.spop(1)
				a = c.spop(1)
				c.setstore(a,b)
				if n > 0:
					if len(c.stack[0]) < n:
						c.stack[1].extend(c.stack[0])
						c.stack[1].extend([0]*(n - len(c.stack[0])))
					else:
						c.stack[1].extend(c.stack[0][-n:])
				elif n < 0:
					del c.stack[1][n:]
				del c.stack[0]
		elif inst == "u": # stack under stack
			if len(c.stack) < 2:
				c.setdelta(-c.dx,-c.dy) # reflect on failure
			else:
				n = c.spop(0)
				if n > 0:
					for i in range(n):
						c.stack[0].append(c.spop(1))
				elif n < 0:
					for i in range(abs(n)):
						c.stack[1].append(c.spop(0))
		elif inst == "g": # get
			b = c.spop(0) + c.sy
			a = c.spop(0) + c.sx
			s = ord(program[b % fungesize[1]][a % fungesize[0]])
			c.stack[0].append(s)
		elif inst == "p": # put
			b = c.spop(0) + c.sy
			a = c.spop(0) + c.sx
			s = chr(c.spop(0) % 256)
			program.putgrid(a,b,s)
			if curses:
				mungechar(b % fungesize[1], a % fungesize[0], s, A_BOLD)
		elif inst == ".": # output decimal
			if len(c.stack[0]) == 0:
				c.setdelta(-c.dx,-c.dy)
			else:
				output_text(str(c.spop(0))+" ")
		elif inst == ",": # output character
			s = chr(c.spop(0) % 256)
			output_text(s)
		elif inst == "&": # input decimal
			s = ''
			char = buffer_read()
			while char not in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '']:
				char = buffer_read()
			if (char == ''):
				c.setdelta(-c.dx,-c.dy)
			else:
				while char in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
					s += char
					char = buffer_read()
				output_text("\n")
				c.stack[0].append(int(s))
		elif inst == "~": # input character
			char = buffer_read()
			if (char == ''):
				c.setdelta(-c.dx,-c.dy)
			else:
				a = ord(char)
				if a == 13: a = 10
				c.stack[0].append(a)
		elif inst == "=": # system execute
			s = c.popstring()
			c.stack[0].append(1) # nope... we didn't even try.
		elif inst == "(": # load fingerprint
			c.setdelta(-c.dx,-c.dy) # fingerprints are not implemented
		elif inst == ")": # unload fingerprint
			c.setdelta(-c.dx,-c.dy) # fingerprints are not implemented
		elif inst == "t": # split
			n = iplist.index(c)
			a = ip(c.x-c.dx,c.y-c.dy,-c.dx,-c.dy,c.sx,c.sy,c.team)
			iplist.insert(n, a)
			ipteams[c.team].add_ip(a)
				
		elif inst == "y": # get info
			n = c.spop(0)
			temp = c.info()
			if n > 0:
				if n >= len(temp):
					c.stack[0].append(c.stack[0][n-len(temp)])
				else:
					c.stack[0].append(c.info()[n-1])
			else:
				temp.reverse()
				c.stack[0].extend(temp)
			del temp
		else:
			a = ord(inst)
			if debug > 1: stderr.write("Unimplemented instruction: "+`a`+" on IP\
#"+`c.id`+", at "+`(c.x,c.y)`+".\n")
	else: # string mode
		if inst == '"': # end stringmode
			c.stringmode = 0
		else:
			s = ord(inst)
			c.stack[0].append(s)

def mungechar(y, x, char, attribs=0):
	try:
		if (ord(char) < 0) or (ord(char) > 254): stdscr.addch(y, x, "~",\
		color_pair(4) | A_BOLD | attribs)
		elif ord(char) == 127: stdscr.addch(y, x, "?", color_pair(4) | A_BOLD | attribs)
		elif char == "z":
			if attribs > 0: a=attribs
			else: a = color_pair(5) | A_BOLD
			stdscr.addch(y, x, "z", a)
		elif ord(char) >= 32: stdscr.addch(y, x, char, attribs)
		else: stdscr.addch(y, x, ord(char)+64, color_pair(4) | A_BOLD | attribs)
	except:
		if debug: stderr.write("Could not show character:"+`(x, y,\
		char, attribs)`)

def redraw():
	if curses:
		for y in range(25):
			for x in range(80):
				mungechar(y, x, program[y][x])
#		for p in iplist:
#			mungechar(p.y, p.x, program[p.y][p.x], A_REVERSE)

def showstack(stack):
	out = `len(stack)` + ":  "
	for i in stack[:80]:
		if 32 <= i < 128:
			out += chr(i)+" "
		else:
			out += "["+str(i)+"] "
	return out


def step(c,inst):
	global toclear
	
	if delay: time.sleep(delay)
	run(c,inst)
	if curses:
		for i in toclear:
			mungechar(i[0], i[1], program[i[0]][i[1]])
		toclear = []
		for p in iplist:
			mungechar(p.y, p.x, program[p.y][p.x], A_REVERSE\
			| color_pair(p.team+1))
			toclear += [(p.y, p.x)]
		try: stdscr.addstr(26, 0,\
		showstack(iplist[0].stack[0])[:79]+'\n', color_pair(1))
		except: pass
		try: stdscr.addstr(27, 0,\
		showstack(iplist[1].stack[0])[:79]+'\n', color_pair(2))
		except: pass
		try: stdscr.addstr(28, 0,\
		output.splitlines()[-1][-79:]+"\n", color_pair(3) | A_BOLD)
		except: pass
		stdscr.refresh()
	# kill runaway pointers
	if len(c.stack[0]) > max_stack:
		team = ipteams[c.team]
		team.del_ip(c)
		del iplist[iplist.index(c)]
		if team.members() == 0:
			if c.team == 0: exit("Green player wins.")
			else: exit("Red player wins.")
	c.findnext()

def main(stdscr):
	global output, ticks, uniqueid, program, iplist, ipteams
	global toclear, delay
	if curses:
		# Define the color pairs
		init_pair(1, COLOR_RED, COLOR_BLACK)
		init_pair(2, COLOR_GREEN, COLOR_BLACK)
		init_pair(3, COLOR_YELLOW, COLOR_BLACK)
		init_pair(4, COLOR_BLUE, COLOR_BLACK)
		init_pair(5, COLOR_BLACK, COLOR_BLACK)
	
		curs_set(0) # make cursor invisible
	
	output = ""
	ticks = 0
	uniqueid = 0
	if len(argv) > 3: delay = int(argv[3])/1000.0
	elif curses: delay = 0.10
	else: delay = 0
	
	#TODO: update for more than 2 programs (could be messy)
	
	startpos = [None, None]
	sizes = [None, None]
	sources = [None, None]

	for i in [1, 2]:
		if os.access(argv[i],os.R_OK) == 0:
			exit("Can't read " + argv[i])
		file = open(argv[i])
		sources[i-1] = map(chomp,file.readlines())
		file.close()
		sizes[i-1] = findsize(sources[i-1])
	
	startpos[0] = (randint(0,fungesize[0]-1), randint(0,fungesize[1]-1))
	
	# find startpos[1]
	topleft = (startpos[0][0] - sizes[1][0], startpos[0][1] - sizes[1][1])
	filled_space = (sizes[0][0] + sizes[1][0], sizes[0][1] + sizes[1][1])
	possible_pos = []
	for y in range(fungesize[1]):
		for x in range(fungesize[0]):
			if (y - topleft[1]) % 80 < filled_space[1] or\
			(x - topleft[0]) % 80 < filled_space[0]:
				pass
			else: possible_pos += [(x, y)]
	
	if possible_pos == []:
		exit("These programs do not fit in Funge-Space.")
	
	startpos[1] = choice(possible_pos)
	del possible_pos		
	
	program = fungespace(fungesize[0],fungesize[1])
	for i in [0,1]:
		program.loadat(sources[i], startpos[i], sizes[i])

	iplist = [ip(startpos[0][0]-1,startpos[0][1],1,0,startpos[0][0],startpos[0][1],0)]
	iplist +=[ip(startpos[1][0]-1,startpos[1][1],1,0,startpos[1][0],startpos[1][1],1)]
	iplist[0].findnext()
	iplist[1].findnext()

	ipteams = [ipteam(iplist[i]) for i in range(progs)]
	
	toclear = []
	redraw()
	
	while 1: # main loop
		ticks += 1
		for team in ipteams:
			curr_ip = team.next_ip()
			instruct = program[curr_ip.y][curr_ip.x]
			step(curr_ip,instruct)


if __name__=='__main__':
	if "-n" in argv:
		del argv[argv.index("-n")]
		curses = 0
		try: main(None)
		except SystemExit, detail:
			if detail: print detail
		except:
			if debug: traceback.print_exc()
	else:
		curses = 1
		try:
			# Initialize curses
			stdscr=initscr()
			start_color()
			noecho() ; cbreak() 
			stdscr.keypad(1)
			
			main(stdscr)  # Enter the main loop
			
			stdscr.keypad(0)
			echo() ; nocbreak()
			endwin()  # Terminate curses
		except SystemExit, detail:
			stdscr.keypad(0)
			echo() ; nocbreak()
			endwin() 
			if detail: print detail
		except:
			# Restore the terminal to a sane state
			stdscr.keypad(0)
			echo() ; nocbreak()
			endwin() 
			if debug: traceback.print_exc()

